# Create synthetic dataset in a top-level function (times, amps, clusters)


def test_unit_stability():
    pass


def test_feat_cutoff():
    pass


def test_wf_similarity():
    pass


def test_firing_rate_coeff_var():
    pass
